
#include "dosbox.h"

#include "parport.h"
//#include "callback.h"
#include "printer_direct.h"
#include "pic.h"
#include <winspool.h>

namespace {
    auto const kJobHeader = "\x1B%-12345X\x1B\x45"; // \x45 == E
    auto const kJobFooter = "\x1B\x45\x1B%-12345X";
    auto const kDocInvalid = 0u;
    
    bool PrinterWrite(HANDLE const printer, void const* const data, size_t const size) {
        static_assert(sizeof(size_t) == sizeof(DWORD), "Fixup limits.");
        assert(size < std::numeric_limits<DWORD>::max());
        
        char const* remainingData = (char const*)data;
        size_t      remainingSize = size;
        while (remainingSize) {
            DWORD written = std::numeric_limits<DWORD>::max();
            if (!WritePrinter(printer, (LPVOID)remainingData, remainingSize, &written)) {
                LOG_MSG("Failed to write to printer. LastErr = 0x%08x. Skipping buffer."
                       ,::GetLastError());
                return false;
            }
            
            assert(written <= remainingSize);
            remainingSize -= written;
            remainingData += written;
        }
        
        return true;
    }
    
    bool PrinterWrite(HANDLE const printer, char const* const data)
    { return PrinterWrite(printer, data, strlen(data)); }
}

PrinterDirect::PrinterDirect(CommandLine* const cmd, Bitu const nr, Bit8u const initIrq)
                            : CParallel(cmd, nr, initIrq)
                            , _pclWrapperInject(cmd->FindExist("addPCLWrapper", false))
{
    std::string tmp;
    if (cmd->FindStringBegin("timeout:", tmp, false)) {
        if (sscanf(tmp.c_str(), "%u", &_timeout) != 1) {
            LOG_MSG("parallel%d: Invalid timeout parameter.", nr + 1); return;
        }
    }
    
    if (!cmd->FindStringBegin("printer:", tmp, false)) {
        LOG_MSG("parallel%d: No printer path specified.", nr + 1); return;
    }
    
    if (!OpenPrinter((LPSTR)tmp.c_str(), &_printer, nullptr)) {
        LOG_MSG("Unable to access printer: %s\nError: 0x%08x", tmp.c_str(), ::GetLastError());
        return;
    }
    assert(_printer);
    
    InstallationSuccessful = !!_printer;
}

PrinterDirect::~PrinterDirect() {
    if (_docCurrent != kDocInvalid) {
        LOG_MSG("Document in progress. Flushing what we already have...");
        flushDocument();
    }
    
    if (_printer) {
        ClosePrinter(_printer);
        _printer = nullptr;
    }
}

Bitu PrinterDirect::Read_PR () { return _regData; }
Bitu PrinterDirect::Read_COM() { return 0xe0 | (_autoFeed ? 0x02 : 0x00) | (_regCntrl & 0xfd); }
Bitu PrinterDirect::Read_SR () { return 0xDF; }

void PrinterDirect::Write_PR (Bitu val) { _regData = val; }

void PrinterDirect::Write_CON(Bitu val) {
    //LOG_MSG("PrinterDirect::Write_CON CS:IP %8x:%8x",SegValue(cs),reg_eip);
    // init printer if bit 4 is switched on
    
    // data is strobed to the parallel printer on the falling edge of strobe bit
    if (!(val & 0x1) && (_regCntrl & 0x1))
        charPrint(_regData);
    
    _regCntrl = val;
    _autoFeed = !!(val & 0x02);
}

void PrinterDirect::Write_IOSEL(Bitu val) {
    LOG(LOG_DOSMISC, LOG_NORMAL)("PrinterDirect::Write_IOSEL - ignored val=0x%08x", val);
}

void PrinterDirect::handleUpperEvent(Bit16u) {
    auto const timeoutExpire = _lastUsedTick + _timeout;
    if (timeoutExpire >= PIC_Ticks) {
        // Port has been touched in the meantime, try again later
        setEvent(0, float(timeoutExpire - PIC_Ticks) + 1);
        return;
    }
    
    flushDocument();
}

bool PrinterDirect::Putchar(Bit8u val) {
    Write_CON(0xD4);
    // strobe data out
    Write_PR(val);
    Write_CON(0xD5); // strobe pulse
    Write_CON(0xD4); // strobe off
    Read_SR();		 // clear ack

#if PARALLEL_DEBUG
    log_par(dbg_putchar, "putchar  0x%2x", val);
    if (dbg_plainputchar) fprintf(debugfp, "%c", val);
#endif

    return true;
}

void PrinterDirect::charPrint(Bit8u c) {
    _lastUsedTick = PIC_Ticks;
    
    if (_docCurrent == kDocInvalid) {
        DOC_INFO_1 info = {0};
        info.pDocName  = "dbDOS Direct Print Document";
        info.pDatatype = "RAW";
        _docCurrent  = StartDocPrinter(_printer, 1, (LPBYTE)&info);
        if (_docCurrent == kDocInvalid) {
            LOG_MSG("Failed to begin printer document. LastErr = 0x%08x", ::GetLastError());
            return;
        }
        /*if (!StartPagePrinter(_printer)) {
            LOG_MSG("Failed to begin new page. LastErr = 0x%08x. Attempting to continue..."
                   ,::GetLastError());
        }*/
        if (_pclWrapperInject && (!PrinterWrite(_printer, kJobHeader)))
            LOG_MSG("Failed to write job header. LastErr = 0x%08x. Attempting to continue..."
                   ,::GetLastError());
        setEvent(0, _timeout);
    }
    
#if PARALLEL_DEBUG
    LOG_MSG("Dump: 0x%02x (%1c)", (unsigned int)c, c);
#endif
    
    PrinterWrite(_printer, &c, sizeof(c));
}

void PrinterDirect::flushDocument() {
    if (_docCurrent == kDocInvalid) return;
    _docCurrent = kDocInvalid;
    
    if (_pclWrapperInject && (!PrinterWrite(_printer, kJobFooter)))
        LOG_MSG("Failed to write job footer. LastErr = 0x%08x. Attempting to continue..."
               ,::GetLastError());
    
    if (!EndDocPrinter (_printer)) LOG_MSG("EndDocPrinter failed. LastErr = 0x%08x"
                                          ,::GetLastError());
}
